# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
import time
from typing import Any
from typing import Callable

import aorta


__all__: list[str] = [
    'test_accept',
    'test_envelope',
    'test_envelope_with_correlation_id',
    'test_envelope_with_audience',
    'test_envelope_with_audiences',
    'test_qualname',
    'test_typecheck',
    'test_is_valid',
    'test_not_is_valid',
    'test_serialize',
    'test_parse_header',
    'test_parse_unknown',
    'test_parse_known',
    'test_parse_valid_unknown',
    'test_global_parse_unknown',
    'test_global_parse_known',
    'test_global_parse_valid_unknown',
    'test_message',
]


def test_accept(message: aorta.types.Publishable):
    envelope = message.envelope()
    envelope.accept()
    assert envelope.metadata.delivery_count == 1


def test_envelope(message: aorta.types.Publishable):
    envelope = message.envelope()
    assert envelope.metadata.correlation_id is not None


def test_envelope_with_correlation_id(message: aorta.types.Publishable):
    e1 = message.envelope()
    e2 = message.envelope(correlation_id=e1.metadata.correlation_id)
    assert e2.metadata.correlation_id == e1.metadata.correlation_id


def test_envelope_with_audience(message: aorta.types.Publishable):
    e1 = message.envelope(audience={'foo'})
    assert e1.metadata.audience == {'foo'}, e1.metadata.audience


def test_envelope_with_audiences(message: aorta.types.Publishable):
    e1 = message.envelope(audience={'foo', 'bar'})
    assert e1.metadata.audience == {'foo', 'bar'}, e1.metadata.audience


def test_is_valid(message: aorta.types.Publishable):
    envelope = message.envelope()
    assert envelope.is_valid()


def test_not_is_valid(message: aorta.types.Publishable):
    envelope = message.envelope()
    time.sleep(2)
    envelope.metadata.ttl = 1
    assert not envelope.is_valid()


def test_message(
    message: aorta.types.Publishable
):
    envelope = message.envelope()
    assert envelope.message == message


def test_qualname(message: aorta.types.Publishable):
    envelope = message.envelope()
    version, kind = envelope.qualname
    assert version == 'v1'
    assert kind == type(message).__name__


def test_serialize(message: aorta.types.Publishable):
    e1 = message.envelope()
    data = bytes(e1)
    e2 = type(e1).model_validate(json.loads(data))
    assert e1.metadata.uid == e2.metadata.uid, f'{e1} != {e2}'


def test_typecheck(
    message: aorta.types.Publishable,
    typecheck: Callable[[aorta.types.MessageHeader], bool]
):
    assert typecheck(message.envelope())


def test_parse_header(
    message: aorta.types.Publishable
):
    envelope = message.envelope()
    assert envelope.metadata.uid == envelope.header.metadata.uid


def test_parse_unknown(
    parse: Callable[[Any], aorta.types.Envelope[Any] | None]
):
    assert parse({'apiVersion': 'v1', 'kind': 'Unknown'}) is None
    assert parse([]) is None
    assert parse(None) is None


def test_parse_known(
    message: aorta.types.Publishable,
    parse: Callable[[Any], aorta.types.Envelope[Any] | None]
):
    e1 = message.envelope()
    e2 = parse(e1.model_dump())
    assert e2 is not None
    assert e1.metadata.uid == e2.metadata.uid


def test_parse_valid_unknown(
    message: aorta.types.Publishable,
    parse: Callable[[Any], aorta.types.Envelope[Any] | None]
):
    cls = type(type(message))
    e2 = parse({
        **message.envelope().model_dump(),
        'apiVersion': 'v1',
        'kind': 'Unknown',
        'type': cls.typename
    })
    assert e2 is not None
    assert not e2.is_known()
    assert e2.kind == 'Unknown', e2.kind
    assert e2.api_version == 'v1', e2.api_version


def test_global_parse_unknown():
    assert aorta.parse({'apiVersion': 'v1', 'kind': 'Unknown'}) is None
    assert aorta.parse([]) is None
    assert aorta.parse(None) is None


def test_global_parse_known(
    message: aorta.types.Publishable,
):
    e1 = message.envelope()
    e2 = aorta.parse(e1.model_dump())
    assert e2 is not None
    assert e1.metadata.uid == e2.metadata.uid


def test_global_parse_valid_unknown(
    message: aorta.types.Publishable,
):
    cls = type(type(message))
    e2 = aorta.parse({
        **message.envelope().model_dump(),
        'apiVersion': 'v1',
        'kind': 'Unknown',
        'type': cls.typename
    })
    assert e2 is not None
    assert not e2.is_known()
    assert e2.kind == 'Unknown', e2.kind
    assert e2.api_version == 'v1', e2.api_version


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

from .types import Envelope
from .types import Message
from .messagehandler import MessageHandler


T = TypeVar('T', bound=Message)


class Sewer(MessageHandler[T]):
    """A :class:`MessageHandler` implementation that is invoked for each
    incoming message, regardless of the type.
    """
    __module__: str = 'aorta'

    async def handle(self, message: Envelope[Any]) -> None:
        raise NotImplementedError

    async def run(self, envelope: Envelope[Any]) -> tuple[bool, Any]:
        result: Any = NotImplemented
        success = False
        try:
            result = await self.handle(envelope)
            success = True
        except Exception as e: # pragma: no cover
            await self.on_exception(e)
            raise
        return success, result

    async def wants(self, message: Any) -> bool: # type: ignore
        return True


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any

from .types import Envelope


class NullTransport:
    logger: logging.Logger = logging.getLogger('canonical')
    __module__: str = 'aorta'
    _sent: list[Envelope[Any]]

    def __init__(self) -> None:
        self._sent = []

    @property
    def sent(self) -> list[Envelope[Any]]:
        return list(self._sent)

    async def send(
        self,
        messages: list[Envelope[Any]],
        is_retry: bool = False
    ) -> None:
        self.logger.warning("NullTransport does not send any messages.")
        self._sent.extend(messages)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import aorta


__all__: list[str] = [
    'test_publish_message',
    'test_retry_message'
]


@pytest.mark.asyncio
async def test_publish_message(
    publisher: aorta.MessagePublisher,
    message: aorta.types.Publishable
):
    await publisher.send([message.envelope()])


@pytest.mark.asyncio
async def test_retry_message(
    publisher: aorta.MessagePublisher,
    message: aorta.types.Publishable
):
    await publisher.send([message.envelope()], is_retry=True)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import AsyncIterable

import pytest
import pytest_asyncio

import aorta


@pytest.fixture # type: ignore
def publisher(
    transport: aorta.types.ITransport
) -> aorta.MessagePublisher:
    return aorta.MessagePublisher(transport=transport)


@pytest.fixture
def runner(
    publisher: aorta.types.IPublisher
) -> aorta.types.IRunner:
    return aorta.LocalRunner(publisher) # type: ignore


@pytest.fixture
def transport() -> aorta.types.ITransport:
    return aorta.NullTransport()


@pytest_asyncio.fixture # type: ignore
async def transaction(
    publisher: aorta.types.IPublisher
) -> AsyncIterable[aorta.types.ITransaction]:
    async with publisher.begin() as tx:
        yield tx
        tx.rollback()


@pytest.fixture # type: ignore
def CommandHandler() -> type[aorta.MessageHandler[Any]]:
    return FooHandler


@pytest.fixture # type: ignore
def EventListener() -> type[aorta.MessageHandler[Any]]:
    return FooListener


@pytest.fixture # type: ignore
def Sewer() -> type[aorta.MessageHandler[Any]]:
    return FallbackHandler


@pytest.fixture
def command() -> aorta.Command:
    return FooCommand(foo=1)


@pytest.fixture
def event() -> aorta.Event:
    return FooEvent(foo=1)


class FooCommand(aorta.Command):
    foo: int


class FooHandler(aorta.CommandHandler[FooCommand]):

    async def handle(self, command: FooCommand):
        return 'foo'


class FallbackHandler(aorta.Sewer):
    pass


class FooEvent(aorta.Event):
    foo: int


class FooListener(aorta.EventListener[FooEvent]):

    async def handle(self, event: FooEvent):
        return 'foo'


aorta.register(FooHandler)
aorta.register(FooListener)
aorta.register(FallbackHandler)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable

import pytest

import aorta
from aorta.types.test.messagetype import *
from aorta.types.test.messagehandler import *
from aorta.types.test.transaction import *
from ..conftest import FooCommand


@pytest.fixture # type: ignore
def typecheck() -> Callable[[aorta.types.MessageHeader], bool]:
    return aorta.types.MessageHeader.is_command


@pytest.fixture # type: ignore
def parse() -> Callable[[Any], aorta.types.Envelope[Any] | None]:
    return aorta.types.CommandType.parse # type: ignore


@pytest.fixture # type: ignore
def MessageHandler(CommandHandler: type[aorta.MessageHandler]) -> type[aorta.MessageHandler]:
    return CommandHandler


@pytest.fixture
def message() -> aorta.types.Publishable:
    return FooCommand(foo=1)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
import types
import typing
from typing import Any

from .command import Command
from .event import Event



class MessageHandlerType(type):
    __module__: str = 'aorta.types'
    handles: type[Command | Event]
    parameter_name: str

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **params: Any
    ) -> 'MessageHandlerType':
        # Do some checks to ensure that handlers are properly implemented.
        is_abstract = namespace.pop('__abstract__', False)
        new_class = super().__new__(cls, name, bases, {**namespace, 'handles': []})
        if is_abstract:
            return new_class
        
        # Ensure that the handle() method accepts the message as the first
        # parameter.
        sig = inspect.signature(new_class.handle) # type: ignore
        parameters = list(sig.parameters.values())
        if len(parameters) < 2:
            raise TypeError(
                f"Invalid number of arguments for {name}.handle(). "
                f"Ensure that the parameters accepted by this method are at "
                f"least {name}.handle(self, {cls.parameter_name}: "
                f"{cls.handles.__name__})."
            )

        arg = parameters[1]
        if arg.name != cls.parameter_name:
            raise TypeError(
                f'The first positional argument to {name}.handle() '
                f'must be named `{cls.parameter_name}`, got `{arg.name}`.'
            )

        args: set[type] = set()
        handles: set[type[Command | Event]] = set()
        for b in getattr(new_class, '__orig_bases__', []):
            args.update(typing.get_args(b))
        for t in args:
            if typing.get_origin(t) in (typing.Union, types.UnionType):
                handles.update(typing.get_args(t))
                continue
            if not inspect.isclass(t) or not issubclass(t, (Command, Event)):
                continue
            handles.add(t)

        new_class.handles = list(handles) # type: ignore
        return new_class


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import aorta
from aorta.types.test.transport import *
from ...conftest import FooEvent


@pytest.fixture
def message() -> aorta.types.Publishable:
    return FooEvent(foo=1)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import aorta


__all__: list[str] = [
    'test_rollback',
    'test_exception_rolls_back',
    'test_transaction_sets_correlation_id',
]


def test_rollback(
    message: aorta.types.Publishable,
    publisher: aorta.types.IPublisher        
):
    tx = aorta.Transaction(publisher=publisher)
    tx.publish(message)
    assert len(tx.pending()) == 1
    tx.rollback()
    assert len(tx.pending()) == 0


@pytest.mark.asyncio
async def test_exception_rolls_back(
    message: aorta.types.Publishable,
    publisher: aorta.types.IPublisher        
):
    tx = aorta.Transaction(publisher=publisher)
    try:
        async with tx:
            tx.publish(message)
            assert len(tx.pending()) == 1
            raise Exception
    except Exception:
        pass
    assert len(tx.pending()) == 0


@pytest.mark.asyncio
async def test_transaction_sets_correlation_id(
    message: aorta.types.Publishable,
    transport: aorta.NullTransport,
    publisher: aorta.types.IPublisher        
):
    tx = aorta.Transaction(publisher=publisher)
    async with tx:
        tx.publish(message)
    assert len(transport.sent) == 1
    assert transport.sent[0].metadata.correlation_id == tx.correlation_id


# Copyright (C) 2020-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
import logging
from contextlib import AsyncExitStack
from typing import get_args
from typing import get_origin
from typing import Annotated
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Generator
from typing import TypeVar

import fastapi
import fastapi.params
from fastapi.dependencies.models import Dependant
from fastapi.dependencies.utils import get_dependant
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from fastapi.dependencies.utils import solve_dependencies

import aorta
from aorta.types import Acknowledgable
from aorta.types import Command
from aorta.types import Event
from aorta.types import Envelope
from aorta.types import IMessageHandler
from aorta.types import IPublisher
from aorta.types import IRunner
from aorta.types import MessageMetadata
from aorta.transaction import Transaction


E = TypeVar('E', bound=Command|Event)


class BaseRunner(IRunner):
    __module__: str = 'aorta.runners'
    dependency_cache: dict[tuple[Callable[..., Any], tuple[str]], Any]
    logger: logging.Logger = logging.getLogger('canonical')
    request: fastapi.Request

    @classmethod
    def inject(cls) -> Any:
        return fastapi.Depends(cls)

    def begin(self, publisher: IPublisher, metadata: MessageMetadata):
        return Transaction(publisher, metadata, metadata.correlation_id) 

    def initialize_handler(
        self,
        transaction: Transaction,
        handler_class: type[IMessageHandler],
        envelope: aorta.types.Envelope[Any]
    ) -> IMessageHandler:
        return handler_class(
            publisher=transaction,
            metadata=envelope.metadata
        )

    async def run(
        self,
        publisher: IPublisher,
        frame: Acknowledgable,
        envelope: Envelope[Any],
        handlers: Iterable[type[IMessageHandler]],
        request: fastapi.Request | None = None
    ) -> None:
        raise NotImplementedError

    async def run_handler(
        self,
        publisher: IPublisher,
        frame: Acknowledgable,
        envelope: Envelope[Any],
        handler_class: type[IMessageHandler],
        request: fastapi.Request | None = None
    ) -> tuple[bool, Any]:
        result = None
        async with AsyncExitStack() as stack:
            cache:  dict[tuple[Callable[..., Any], tuple[str]], Any] = {}
            async with self.begin(publisher, envelope.metadata) as tx:
                handler = self.initialize_handler(tx, handler_class, envelope)
                dependant, kwargs = await self.inject_dependencies(
                    stack=stack,
                    handler=handler,
                    envelope=envelope,
                    request=request,
                    cache=cache
                )

                assert dependant.call is not None
                assert callable(dependant.call)
                result = await dependant.call(**kwargs)
        return True, result

    async def inject_dependencies(
        self,
        stack: AsyncExitStack,
        handler: IMessageHandler,
        envelope: Envelope[Any],
        cache: dict[tuple[Callable[..., Any], tuple[str]], Any],
        request: fastapi.Request | None = None
    ):
        dependant = get_dependant(call=handler.run, path='/')
        dependant.dependencies.extend(self.get_injectors(handler))
        values, errors, *_ = await solve_dependencies(
            request=self.request_factory(request),
            dependant=dependant,
            body=envelope.model_dump(),
            dependency_overrides_provider=None,
            dependency_cache=cache,
            async_exit_stack=stack
        )
        if errors:
            raise ValueError(errors)
        cache.update(cache)
        assert dependant.call is not None
        assert callable(dependant.call)
        values['envelope'] = envelope
        return dependant, values

    def get_injectors(self, obj: IMessageHandler) -> Generator[Dependant, None, None]:
        for attname, member in inspect.getmembers(obj):
            if not isinstance(member, fastapi.params.Depends):
                continue
            def setter(attname: str, dep: Any = member):
                async def f(dep: Any = dep) -> None:
                    setattr(obj, attname, dep)
                return f
            yield get_parameterless_sub_dependant(
                 depends=fastapi.Depends(setter(attname, member)),
                 path='/'
            )

        seen: set[str] = set()
        for cls in inspect.getmro(type(obj)):
            annotations = inspect.get_annotations(cls)
            for attname, annotation in annotations.items():
                if attname in seen:
                    continue
                if get_origin(annotation) != Annotated:
                    continue
                args = get_args(annotation)
                if len(args) != 2 or not isinstance(args[1], fastapi.params.Depends):
                    continue

                def setter(attname: str, dep: Any):
                    async def f(dep: Any = dep) -> None:
                        setattr(obj, attname, dep)
                    return f
                yield get_parameterless_sub_dependant(
                    depends=fastapi.Depends(setter(attname, args[1])),
                    path='/'
                )
                seen.add(attname)

    def request_factory(
        self,
        request: fastapi.Request | None
    ) -> fastapi.Request:
        return request or fastapi.Request({
            'type': 'http',
            'query_string': '',
            'headers': []
        })


# Copyright (C) 2016-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Any

from fastapi import Request
from google.cloud.pubsub_v1 import SubscriberClient # type: ignore
from google.cloud.pubsub_v1.subscriber.exceptions import AcknowledgeError # type: ignore
from google.cloud.pubsub_v1.subscriber.futures import StreamingPullFuture # type: ignore
from google.cloud.pubsub_v1.subscriber.message import Message  # type: ignore
from google.cloud.pubsub_v1.types import FlowControl  # type: ignore
from libcanonical.utils.logging import LoggingConfigDict # type: ignore

import aorta
from aorta import MessagePublisher
from aorta.ext.consumer import BaseConsumer
from aorta.runners import SequentialRunner
from aorta.types import Acknowledgable
from aorta.types import IRunner
from aorta.types import Envelope
from aorta.types import IPublisher
from aorta.types import ITransport

from .googletransport import GoogleTransport


class TopicListener(BaseConsumer):
    debug = True
    project: str | None = None
    publisher: IPublisher
    streamer: StreamingPullFuture
    transport_class: type[GoogleTransport] = GoogleTransport

    def acknowledge(self, message: Acknowledgable) -> None:
        message.ack()

    def accept_message(self, message: Message):
        envelope = None
        try:
            envelope = aorta.loads(message.data)
            if envelope is None:
                raise ValueError("Unable to parse message from incoming data.")
        except Exception as e:
            message.ack()
            if self.debug:
                self.logger.debug("Caught fatal %s", repr(e))
            self.logger.warning("Received a malformed message.")
            return
        assert envelope is not None
        if isinstance(envelope, Envelope):
            self.queue.put((message, envelope)) # type: ignore

    def configure(self, reloading: bool = False):
        self.runner = SequentialRunner()
        if self.project is None:
            self.project = os.environ['GOOGLE_PUBSUB_PROJECT']
        super().configure(reloading=reloading)
        if not reloading:
            self.publisher = self.get_publisher()
            self.subscriber = SubscriberClient()
            self.streamer = self.subscriber.subscribe( # type: ignore
                os.environ['GOOGLE_PUBSUB_SUBSCRIPTION'],
                self.accept_message
            )

    def get_flow_control(self) -> FlowControl:
        return FlowControl(max_messages=self.concurrency)

    def get_logging_config(self) -> LoggingConfigDict:
        config = super().get_logging_config()
        config['loggers']['aortra.metrics'] = {
            'handlers': ['google-cloud'],
            'level': 'INFO',
            'propagate': False
        }
        return config

    def get_publisher(self) -> IPublisher:
        return MessagePublisher(transport=self.get_transport())

    def get_runner(
        self,
        publisher: IPublisher,
        frame: Acknowledgable,
        request: Request | None = None,
        **kwargs: Any
    ) -> IRunner:
        return self.runner

    def route(self, envelope: Envelope[Any]) -> list[str]:
        if not isinstance(envelope.message, aorta.Command):
            return []
        return ['dataverse.commands']

    async def teardown(self) -> None:
        await super().teardown()
        self.streamer.cancel()

    def get_transport(self) -> ITransport:
        assert self.project is not None
        return self.transport_class(
            project=self.project,
            prefix=self.name
        )

    async def main_event(self) -> None:
        try:
            self.streamer.result(timeout=1.0)
        except TimeoutError:
            pass
        return await super().main_event()


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic
import pydantic.main

from .eventenvelope import EventEnvelope
from .messageheader import MessageHeader
from .messagetype import MessageType


class UnknownEventEnvelope(EventEnvelope[Any]):
    data: dict[str, Any]

    def is_known(self) -> bool:
        return False


class EventType(MessageType):
    __module__: str = 'aorta.types'
    __registry__: dict[tuple[str, str], type[EventEnvelope[Any]]] = {}
    envelope_attr: str = 'data'
    envelope_class: type[EventEnvelope[Any]] = EventEnvelope
    typename: str = 'cochise.io/event'

    @staticmethod
    def parse(data: Any) -> EventEnvelope[Any] | MessageHeader | None:
        header = None
        try:
            header = MessageHeader.model_validate(data)
            if header.type == EventType.typename:
                return EventType.__registry__[(header.api_version, header.kind)].model_validate(data)
        except KeyError:
            return UnknownEventEnvelope.model_validate(data)
        except (pydantic.ValidationError, TypeError, ValueError):
            return header


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

from .envelope import Envelope
from .itransaction import ITransaction
from .message import Message


class IPublisher(Protocol):
    __module__: str = 'aorta.types'

    def begin(self) -> ITransaction: ...

    async def publish(
        self,
        message: Message,
        **kwargs: Any
    ) -> None:
        ...

    async def send(
        self,
        messages: list[Envelope[Any]],
        is_retry: bool = False
    ): ...


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol
from typing import TypeVar

from .publishable import Publishable


T = TypeVar('T', bound='ITransaction')


class ITransaction(Protocol):
    __module__: str = 'aorta.types'
    correlation_id: str

    def rollback(self) -> None:
        ...

    def pending(self) -> list[Publishable]:
        ...

    def publish(
        self,
        message: Publishable,
        correlation_id: str | None = None,
        audience: set[str] | None = None
    ):
        ...

    async def commit(self) -> None: ...
    async def send(self, message: Publishable) -> None: ...
    async def __aenter__(self: T) -> T: ...
    async def __aexit__(self, cls: type[BaseException] | None, *args: Any): ...


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeVar

import pydantic

from .commandtype import CommandType
from .envelopedmixin import EnvelopeMixin


T = TypeVar('T', bound='Command')


class Command(pydantic.BaseModel, EnvelopeMixin, metaclass=CommandType):
    __message_attr__ = 'spec'
    __message_type__ = 'cochise.io/command'


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Iterable

import fastapi

from .acknowledgable import Acknowledgable
from .envelope import Envelope
from .imessagehandler import IMessageHandler
from .ipublisher import IPublisher


class IRunner:
    __module__: str = 'aorta.types'

    async def run(
        self,
        publisher: IPublisher,
        frame: Acknowledgable,
        envelope: Envelope[Any],
        handlers: Iterable[type[IMessageHandler]],
        request: fastapi.Request | None = None
    ) -> Any:
        ...


# Copyright (C) 2020-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Annotated
from typing import Any
from typing import TypeAlias

import fastapi

import aorta
from aorta.types import IPublisher
from .messagetransport import MessageTransport


class FastAPIMessagePublisher(aorta.MessagePublisher):
    __module__: str = 'aorta.ext.fastapi'
    logger: logging.Logger = logging.getLogger('aorta')
    request: fastapi.Request

    def __init__(self, transport: MessageTransport):
        super().__init__(transport=transport)

    async def send(
        self,
        messages: list[aorta.types.Envelope[Any]],
        is_retry: bool = False
    ) -> None:
        for message in messages:
            self.logger.debug(
                "Publishing message %s/%s (id: %s)",
                message.api_version, message.kind, message.metadata.uid
            )
        return await super().send(messages, is_retry)


async def get(request: fastapi.Request):
    return getattr(request.state, 'aorta_publisher')


MessagePublisher: TypeAlias = Annotated[IPublisher, fastapi.Depends(FastAPIMessagePublisher)]


# Copyright (C) 2016-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any
#from typing import Callable
from queue import SimpleQueue

#import fastapi

from aorta.types import Acknowledgable
from aorta.types import Envelope
from aorta.types import IRunner
#from aorta.types import IPublisher


class Pool:
    logger: logging.Logger = logging.getLogger('canonical')
    runner: IRunner

    def __init__(
        self,
        queue: SimpleQueue[tuple[Acknowledgable, Envelope[Any]]],
        concurrency: int = 1,
        debug: bool = False
    ) -> None:
        self.concurrency = concurrency
        self.debug = debug
        self.queue = queue

    async def join(self) -> None:
        raise NotImplementedError

    #async def submit(
    #    self,
    #    ack: Callable[..., None],
    #    runner: IRunner,
    #    publisher: IPublisher,
    #    frame: Acknowledgable,
    #    envelope: Envelope[Any],
    #    request: fastapi.Request | None = None
    #) -> None:
    #    raise NotImplementedError


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import aorta
from aorta.ext.google import PubsubTransport



@pytest.fixture
def transport() -> aorta.types.ITransport:
    return PubsubTransport(
        project='unimatrixdev',
        topic='aorta.events',
        retry_topic='aorta.retrying'
    )


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any
from typing import Generic
from typing import TypeVar

import pydantic

from .types import Command
from .types import Drop
from .types import Envelope
from .types import Event
from .types import IMessageHandler
from .types import ITransaction
from .types import MessageMetadata
from .types import Publishable


D = TypeVar('D')
T = TypeVar('T', bound=Command | Event)


class MessageHandler(IMessageHandler, Generic[T]):
    """The base class for all message handlers."""
    __module__: str = 'aorta'
    __abstract__: bool = True
    handles: list[type[Command] | type[Event]]
    logger: logging.Logger = logging.getLogger('canonical')
    metrics: logging.Logger = logging.getLogger('aorta.metrics')
    metadata: MessageMetadata
    publisher: ITransaction

    def __init__(
        self,
        publisher: ITransaction,
        metadata: MessageMetadata
    ):
        self.metadata = MessageMetadata.model_validate(metadata.model_dump())
        self.publisher = publisher
        assert metadata.correlation_id == self.metadata.correlation_id
        assert metadata == self.metadata

    def issue(self, command: Command, audience: set[str] | None = None):
        """Issue a command using the default command issuer."""
        assert self.metadata.correlation_id is not None
        self.publisher.publish(
            command,
            correlation_id=self.metadata.correlation_id,
            audience=audience
        )

    def label(self, name: str, decoder: type[D] = str) -> D | None:
        return self._get_mapping(self.metadata.labels, name, decoder)

    async def on_exception(self, exception: Exception) -> None:
        """Hook to perform cleanup after a fatal exception.
        """
        pass

    def publish(self, event: Event, audience: set[str] | None = None):
        """Publish an event using the default event publisher."""
        assert self.metadata.correlation_id is not None
        self.publisher.publish(
            event,
            correlation_id=self.metadata.correlation_id,
            audience=audience
        )

    async def commit(self):
        await self.publisher.commit()

    async def run(self, envelope: Envelope[Any]) -> tuple[bool, Any]:
        result: Any = NotImplemented
        success = False
        try:
            # Validate the message again to create a new object, to prevent
            # unstable behavior if handlers modify the message.
            message = envelope.message.model_validate(envelope.message)
            result = await self.handle(message) # type: ignore
            success = True
        except Drop:
            success = True
        except Exception as e: # pragma: no cover
            cls = type(self)
            self.logger.exception(
                'Caught fatal %s in handler %s: %s',
                type(e).__name__,
                f'{cls.__module__}.{cls.__name__}',
                result
            )
            await self.on_exception(e)
            raise
        return success, result

    async def send(self, message: Publishable) -> None:
        """Immediately send the given message."""
        return await self.publisher.send(message)

    async def wants(self, _: Any) -> bool:
        raise NotImplementedError

    def _get_mapping(
        self,
        mapping: dict[str, Any],
        name: str,
        decoder: type[D] = str
    ) -> D | None:
        if name not in mapping:
            return None
        a = pydantic.TypeAdapter(decoder)
        return a.validate_python(mapping[name])